# -*- coding: utf-8 -*-

from imio.email.dms.utils import get_next_id
from imio.email.dms.utils import get_reduced_size
from imio.email.dms.utils import reception_date
from imio.email.dms.utils import safe_text
from imio.email.dms.utils import save_attachment
from imio.email.dms.utils import set_next_id
from imio.email.parser.parser import Parser
from imio.email.parser.tests.test_parser import get_eml_message
from pathlib import Path
from unittest.mock import patch

import configparser
import unittest


class TestUtils(unittest.TestCase):
    def test_safe_text(self):
        test_bytes = "\u01b5".encode("utf-8")
        self.assertEqual(safe_text("spam"), "spam")
        self.assertEqual(safe_text(b"spam"), "spam")
        self.assertEqual(safe_text("spam"), "spam")
        self.assertEqual(safe_text("spam".encode("utf-8")), "spam")
        self.assertEqual(safe_text(test_bytes), "\u01b5")
        self.assertEqual(safe_text("\xc6\xb5".encode("iso-8859-1")), "\u01b5")
        self.assertEqual(safe_text(test_bytes, encoding="ascii"), "\u01b5")
        self.assertEqual(safe_text(1), 1)
        self.assertIsNone(safe_text(None))

    def test_save_attachment(self):
        email = get_eml_message("01_email_with_inline_and_annexes.eml")
        attachment = Parser(email, False, "1").attachments[0]
        folder = "/tmp"
        save_attachment(folder, attachment)

        self.assertTrue(Path(f"{folder}/2-1-page-daccueil.png").exists())
        with open("/tmp/2-1-page-daccueil.png", "rb") as file:
            self.assertEqual(file.read(), attachment["content"])

    def test_reception_date(self):
        email = get_eml_message("01_email_with_inline_and_annexes.eml")
        self.assertEqual(reception_date(email), "2025-01-06 07:51")
        email = get_eml_message("02_email_with_inline_annex_eml.eml")
        self.assertEqual(reception_date(email), "")
        email = get_eml_message("03_email_with_false_inline.eml")
        self.assertEqual(reception_date(email), "2022-02-23 15:02")
        email = get_eml_message("04_email_with_pdf_attachment.eml")
        self.assertEqual(reception_date(email), "2025-01-09 11:10")

    def test_get_reduced_size(self):
        self.assertTupleEqual((False, (500, 500)), get_reduced_size((500, 500), (None, None)))
        self.assertTupleEqual((False, (500, 500)), get_reduced_size((500, 500), (600, None)))
        self.assertTupleEqual((True, (400, 400)), get_reduced_size((500, 500), (400, 450)))
        self.assertTupleEqual((True, (400, 333)), get_reduced_size((600, 500), (400, None)))
        self.assertTupleEqual((True, (300, 400)), get_reduced_size((600, 800), (None, 400)))

    def test_next_id(self):
        config = configparser.ConfigParser()
        config.read("../../config.ini")

        devinfos = {"nid": None}
        with patch("imio.email.dms.utils.dev_mode", True):
            set_next_id(config, 0)
            self.assertTupleEqual((1, "01Z9999"), get_next_id(config, devinfos))
            self.assertTupleEqual((2, "01Z9999"), get_next_id(config, devinfos))
            self.assertTupleEqual((3, "01Z9999"), get_next_id(config, devinfos))

        devinfos = {"nid": None}
        with patch("imio.email.dms.utils.dev_mode", False):
            set_next_id(config, 0)
            self.assertTupleEqual((1, "01Z9999"), get_next_id(config, devinfos))
            self.assertTupleEqual((1, "01Z9999"), get_next_id(config, devinfos))
            self.assertTupleEqual((1, "01Z9999"), get_next_id(config, devinfos))



# -*- coding: utf-8 -*-
from imio.email.dms import dev_mode
from imio.email.dms.utils import reception_date
from imio.email.parser import email_policy  # noqa
from imio.email.parser.parser import Parser  # noqa
from mailparser import MailParser

import chardet
import email
import imaplib
import logging
import six


logger = logging.getLogger("imio.email.dms")


class MailData(object):
    def __init__(self, mail_id, mail_obj):
        self.id = mail_id
        self.mail = mail_obj


class IMAPEmailHandler(object):
    """Handle IMAP mails"""

    connection = None

    def connect(self, host, port, ssl, login, password):
        """Connect and login to IMAP server"""
        if ssl:
            self.connection = imaplib.IMAP4_SSL(host, port)
        else:
            self.connection = imaplib.IMAP4(host, port)
        self.connection.login(login, password)
        self.connection.select("INBOX")

    def disconnect(self):
        """Disconnect from IMAP server"""
        self.connection.close()
        self.connection.logout()

    def reset_errors(self):
        """Fetch all messages in error status and put them back in waiting"""
        res, data = self.connection.search(None, u"KEYWORD error")
        if res != "OK":
            logger.error("Unable to fetch mails")
            return []
        amount = 0
        for mail_id in data[0].split():
            amount += 1
            self.mark_reset_error(mail_id)
        return amount

    def get_mail(self, mail_id):
        res, mail_data = self.connection.fetch(mail_id, "(RFC822)")
        if res != "OK":
            logger.error("Unable to fetch mail {0}".format(mail_id))
            return None
        mail_body = mail_data[0][1]
        if six.PY3:
            try:
                mail_body = mail_body.decode("utf-8")
            except UnicodeDecodeError:
                detection = chardet.detect(mail_body)
                mail_body = mail_body.decode(detection["encoding"])
        mail = email.message_from_string(mail_body, policy=email_policy)
        return mail

    def get_waiting_emails(self):
        """Fetch all waiting messages"""
        args = [u"NOT KEYWORD imported", u"NOT KEYWORD unsupported", u"NOT KEYWORD error", u"NOT KEYWORD ignored"]
        # args = [u"KEYWORD imported"]
        # args = ['SUBJECT "PERMANNE"']
        res, data = self.connection.search(None, *args)
        if res != "OK":
            logger.error("Unable to fetch mails")
            return []
        waiting = []
        for mail_id in data[0].split():
            # if mail_id != b'388': continue
            mail = self.get_mail(mail_id)
            if not mail:
                continue
            mail_infos = MailData(mail_id, mail)
            waiting.append(mail_infos)
        return waiting

    def should_handle(self, mail_id):
        res, flags_data = self.connection.fetch(mail_id, "(FLAGS)")
        if res != "OK":
            logger.error("Unable to fetch flags for mail {0}".format(mail_id))
            return False
        flags = imaplib.ParseFlags(flags_data[0])
        if b"imported" in flags or b"error" in flags or b"unsupported" in flags:
            return False
        return True

    def list_last_emails(self, nb=20):
        """List last messages"""
        # args = [u"NOT KEYWORD imported", u"NOT KEYWORD unsupported", u"NOT KEYWORD error", u"NOT KEYWORD ignored"]
        # args = ['SUBJECT "PERMANNE"']
        args = ["ALL"]
        res, data = self.connection.search(None, *args)
        if res != "OK":
            logger.error("Unable to fetch mails")
            return []
        lst = []
        for mail_id in data[0].split()[-nb:]:
            res, flags_data = self.connection.fetch(mail_id, "(FLAGS)")
            if res != "OK":
                logger.error("Unable to fetch flags for mail {0}".format(mail_id))
                continue
            flags = imaplib.ParseFlags(flags_data[0])
            if isinstance(mail_id, bytes):
                mail_id = mail_id.decode()
                flags = [fl.decode() for fl in flags]
            mail = self.get_mail(mail_id)
            if not mail:
                continue
            parser = Parser(mail, dev_mode, mail_id)
            parsed_orig_mail = MailParser(mail)
            r_date = reception_date(mail)
            lst.append(
                u"{}, {}: '{}', '{}', '{}', '{}', {}".format(
                    r_date,
                    mail_id,
                    parsed_orig_mail.headers["From"],
                    parsed_orig_mail.subject,
                    parser.parsed_message.headers.get("From"),
                    parser.headers["Subject"],
                    flags,
                )
            )
            # parser.message.get('message-id'), flags))
            logger.info(lst[-1])
        return lst

    def stats(self):
        """List all flags"""
        res, data = self.connection.search(None, "ALL")
        if res != "OK":
            logger.error("Unable to fetch mails")
            return []
        lst = []
        stats = {"tot": 0, "flags": {}}
        for mail_id in data[0].split():
            stats["tot"] += 1
            res, flags_data = self.connection.fetch(mail_id, "(FLAGS)")
            if res != "OK":
                logger.error("Unable to fetch flags for mail {0}".format(mail_id))
                continue
            for flag in imaplib.ParseFlags(flags_data[0]):
                flag = flag.decode()
                if flag not in stats["flags"]:
                    stats["flags"][flag] = 0
                stats["flags"][flag] += 1
        return stats

    def mark_reset_error(self, mail_id):
        """Reset 'error' / 'waiting' flags on specified mail"""
        self.connection.store(mail_id, "-FLAGS", "imported")
        self.connection.store(mail_id, "-FLAGS", "error")
        self.connection.store(mail_id, "+FLAGS", "waiting")

    def mark_reset_ignored(self, mail_id):
        """Reset 'ignored' / 'waiting' flags on specified mail"""
        self.connection.store(mail_id, "-FLAGS", "imported")
        self.connection.store(mail_id, "-FLAGS", "ignored")
        self.connection.store(mail_id, "+FLAGS", "waiting")

    def mark_reset_all(self, mail_id):
        """Reset all flags on specified mail"""
        self.connection.store(mail_id, "-FLAGS", "imported error ignored unsupported")
        # self.connection.store(mail_id, "+FLAGS", "waiting")

    def mark_mail_as_imported(self, mail_id):
        """(Un)Mark 'imported' / 'waiting' flags on specified mail"""
        self.connection.store(mail_id, "-FLAGS", "waiting")
        self.connection.store(mail_id, "-FLAGS", "error")
        self.connection.store(mail_id, "+FLAGS", "imported")

    def mark_mail_as_error(self, mail_id):
        """(Un)Mark 'error' / 'waiting' flags on specified mail"""
        self.connection.store(mail_id, "-FLAGS", "waiting")
        self.connection.store(mail_id, "+FLAGS", "error")

    def mark_mail_as_unsupported(self, mail_id):
        """(Un)Mark 'unsupported' / 'waiting' flags on specified mail"""
        self.connection.store(mail_id, "-FLAGS", "waiting")
        self.connection.store(mail_id, "+FLAGS", "unsupported")

    def mark_mail_as_ignored(self, mail_id):
        """(Un)Mark 'ignore' / 'waiting' flags on specified mail"""
        self.connection.store(mail_id, "-FLAGS", "waiting")
        self.connection.store(mail_id, "+FLAGS", "ignored")



# -*- coding: utf-8 -*-
import logging


dev_mode = False

logger = logging.getLogger("imio.email.dms")
logger.setLevel(logging.INFO)
chandler = logging.StreamHandler()
chandler.setLevel(logging.INFO)
logger.addHandler(chandler)



# -*- coding: utf-8 -*-
from datetime import datetime
from email import generator
from email import utils
from imio.email.dms import dev_mode

import os
import six


try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path  # noqa


def safe_text(value, encoding="utf-8") -> str:
    """Converts a value to text, even it is already a text string.
    Copied from plone.base.utils
    """
    if isinstance(value, bytes):
        try:
            value = str(value, encoding)
        except UnicodeDecodeError:
            value = value.decode("utf-8", "replace")
    return value


def save_attachment(folder, at_dic):
    file_path = os.path.join(folder, at_dic["filename"])
    with open(file_path, "wb") as file:
        file.write(at_dic["content"])


def save_as_eml(path, message):
    with open(path, "w") as emlfile:
        gen = generator.Generator(emlfile)
        gen.flatten(message)


def reception_date(message):
    """Returns localized mail date"""
    date_str = message.get("date")
    r_date = u""
    if date_str:
        date_tuple = utils.parsedate_tz(date_str)
        if date_tuple:
            date = datetime.fromtimestamp(utils.mktime_tz(date_tuple))
            r_date = date.strftime("%Y-%m-%d %H:%M")
    return r_date


def get_next_id(config, dev_infos):
    """Get next id from counter file"""
    ws = config["webservice"]
    client_id = "{0}Z{1}".format(ws["client_id"][:2], ws["client_id"][-4:])
    counter_dir = Path(ws["counter_dir"])
    next_id_path = counter_dir / client_id
    if next_id_path.exists() and next_id_path.read_text():
        next_id = int(next_id_path.read_text()) + 1
    else:
        next_id = 1
    if dev_mode:
        if dev_infos["nid"] is None:
            dev_infos["nid"] = next_id
        else:
            dev_infos["nid"] += 1
            return dev_infos["nid"], client_id
    return next_id, client_id


def get_reduced_size(size, max_size):
    """Resize an image size while maintaining the aspect ratio.

    :param size: Original size of the image (width, height).
    :param max_size: Maximum size (max_width, max_height).
    :return: A boolean indicating if resizing occurred and the new size (width, height).
    """
    original_width, original_height = size
    max_width, max_height = max_size

    if max_width is None and max_height is None:
        return False, size

    aspect_ratio = original_width / original_height
    new_width, new_height = original_width, original_height

    if max_width is not None and new_width > max_width:
        new_width = max_width
        new_height = int(new_width / aspect_ratio)

    if max_height is not None and new_height > max_height:
        new_height = max_height
        new_width = int(new_height * aspect_ratio)

    size_reduced = new_width != original_width or new_height != original_height
    return size_reduced, (new_width, new_height)


def get_unique_name(filename, files):
    """Get a filename and eventually rename it so it is unique in files list"""
    new_filename = filename
    counter = 1
    filename, extension = os.path.splitext(filename)
    while new_filename in files:
        new_filename = "{} ({}){}".format(filename, counter, extension)
        counter += 1
    files.append(new_filename)
    return new_filename


def set_next_id(config, current_id):
    """Set current id in counter file"""
    ws = config["webservice"]
    client_id = "{0}Z{1}".format(ws["client_id"][:2], ws["client_id"][-4:])
    counter_dir = Path(ws["counter_dir"])
    next_id_path = counter_dir / client_id
    current_id_txt = str(current_id) if six.PY3 else str(current_id).decode()
    next_id_path.write_text(current_id_txt)



# -*- coding: utf-8 -*-
"""Installer for the imio.email.dms package."""

from setuptools import find_packages
from setuptools import setup


long_description = "\n\n".join(
    [
        open("README.rst").read(),
        open("CONTRIBUTORS.rst").read(),
        open("CHANGES.rst").read(),
    ]
)

setup(
    name="imio.email.dms",
    version="0.29.4",
    description="Package to read emails and send them to DMS",
    long_description=long_description,
    # Get more from https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
    ],
    python_requires=">=3.10, <3.11",
    keywords="Python",
    author="Laurent Lasudry",
    author_email="info@affinitic.be",
    url="https://pypi.python.org/pypi/imio.email.dms",
    license="GPL version 2",
    packages=find_packages("src", exclude=["ez_setup"]),
    namespace_packages=["imio", "imio.email"],
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        # -*- Extra requirements: -*-
        "setuptools",
        "docopt",
        "zc.lockfile",
        "imio.email.parser",
        "imio.pyutils",
        'pathlib2;python_version<"3.0"',
        "requests",
        "six",
        "configparser",
    ],
    extras_require={},
    entry_points="""
    [console_scripts]
    process_mails = imio.email.dms.main:process_mails
    clean_mails = imio.email.dms.main:clean_mails
    """,
)



from docopt import docopt
from imio.email.dms.imap import MailData
from imio.email.dms.main import __doc__
from imio.email.dms.main import clean_mails
from imio.email.dms.main import compress_pdf
from imio.email.dms.main import modify_attachments
from imio.email.dms.main import Notify
from imio.email.dms.main import process_mails
from imio.email.dms.main import resize_inline_images
from imio.email.parser import email_policy  # noqa
from imio.email.parser.parser import Parser
from imio.email.parser.tests import test_parser
from imio.email.parser.tests.test_parser import get_eml_message
from pathlib import Path
from unittest.mock import patch

import configparser
import email
import os
import PyPDF2
import tarfile
import unittest


TEST_FILES_PATH = os.path.join(os.path.dirname(__file__), "files")
EML_TEST_FILES_PATH = os.path.join(os.path.dirname(test_parser.__file__), "files")


class TestMain(unittest.TestCase):
    def setUp(self):
        self.config = configparser.ConfigParser()

    def test_modify_attachments(self):
        to_tests = [
            {
                "fn": "01_email_with_inline_and_annexes.eml",
                "orig": {"nb": 4, "len": [269865, 673, 9309, 310852]},
                "mod": {"all_nb": 4, "at_nb": 2, "len": [186946, 673, 9309, 154746], "mod": [True, None, None, True]},
            },
            {
                "fn": "04_email_with_pdf_attachment.eml",
                "orig": {"nb": 1, "len": [2231999]},
                "mod": {"all_nb": 1, "at_nb": 1, "len": [160165], "mod": [True]},
            },
        ]
        # breakpoint()
        for dic in to_tests:
            name = dic["fn"]
            eml = get_eml_message(name)
            parser = Parser(eml, False, name)
            self.assertEqual(len(parser.attachments), dic["orig"]["nb"])
            self.assertListEqual([at["len"] for at in parser.attachments], dic["orig"]["len"])
            mod_attach = modify_attachments(name, parser.attachments)
            self.assertEqual(len(mod_attach), dic["mod"]["all_nb"])
            self.assertListEqual([at["len"] for at in mod_attach], dic["mod"]["len"])
            self.assertListEqual([at.get("modified") for at in mod_attach], dic["mod"]["mod"])
            mod_attach = modify_attachments(name, parser.attachments, with_inline=False)
            self.assertEqual(len(mod_attach), dic["mod"]["at_nb"])

    def test_compress_pdf(self):
        pdf_file = os.path.join(TEST_FILES_PATH, "pdf-example-bookmarks-1-2.pdf")
        with open(pdf_file, "rb") as pdf:
            pdf_content = pdf.read()
        compressed_pdf_content = compress_pdf(pdf_content)
        self.assertGreater(len(pdf_content), len(compressed_pdf_content))

        with open("pdf-example-bookmarks-1-2-compressed.pdf", "wb") as f:
            f.write(compressed_pdf_content)

        with open("pdf-example-bookmarks-1-2-compressed.pdf", "rb") as f:
            reader = PyPDF2.PdfFileReader(f)
            self.assertFalse(reader.getIsEncrypted())
            self.assertEqual(reader.getNumPages(), 2)
            self.assertEqual(reader.getDocumentInfo().get("/Producer"), "GPL Ghostscript 10.02.1")

    def test_resize_inline_images(self):
        def get_html_part(message):
            for part in message.walk():
                if part.get_content_type() == "text/html":
                    return part
            return None

        to_tests = [
            {
                "fn": "01_email_with_inline_and_annexes.eml",
                "orig": {
                    "html_parts": ['<img src="cid:ii_m5kspqrb0" alt="2-1-page-daccueil.png" width="1920" height="953">']
                },
                "mod": {
                    "html_parts": [
                        '<img alt="2-1-page-daccueil.png" height="953" src="cid:ii_m5kspqrb0" style="max-width: 100%; height: auto; width: auto" width="1920"/>'
                    ]
                },
            },
            {
                "fn": "02_email_with_inline_annex_eml.eml",
                "orig": {
                    "html_parts": ['<img src="cid:ii_m5kuur6b1" alt="organization_icon.png" width="16" height="16">']
                },
                "mod": {
                    "html_parts": ['<img src="cid:ii_m5kuur6b1" alt="organization_icon.png" width="16" height="16">']
                },
            },
            {
                "fn": "03_email_with_false_inline.eml",
                "orig": {"html_parts": []},
                "mod": {"html_parts": []},
            },
            {
                "fn": "04_email_with_pdf_attachment.eml",
                "orig": {"html_parts": []},
                "mod": {"html_parts": []},
            },
        ]

        for dic in to_tests:
            mail_name = dic["fn"]
            eml = get_eml_message(mail_name)
            parser = Parser(eml, False, mail_name)

            attachments = modify_attachments(mail_name, parser.attachments, with_inline=True)
            new_message = resize_inline_images(mail_name, parser.message, attachments)
            new_parser = Parser(new_message, False, mail_name, extract=False)

            self.assertEqual(len(new_parser.attachments), len(parser.attachments))
            self.assertGreaterEqual(len(parser.message.as_string()), len(new_message.as_string()))

            for old_at, new_at in zip(attachments, new_parser.attachments):
                if old_at["disp"] != "inline":
                    continue
                self.assertGreaterEqual(old_at["len"], new_at["len"])

            for i in range(len(dic["orig"]["html_parts"])):
                old_html_part = dic["orig"]["html_parts"][i]
                new_html_part = dic["mod"]["html_parts"][i]
                self.assertIn(
                    old_html_part,
                    get_html_part(parser.message).get_content(),
                )
                self.assertIn(
                    new_html_part,
                    get_html_part(new_message).get_content(),
                )

    def test_process_mails_help(self):
        with self.assertRaises(SystemExit):
            docopt(__doc__, argv=["--help"])

    @patch("imio.email.dms.main.dev_mode", True)
    def test_process_mails_test_eml(self):
        to_test = [
            {
                "filename": "01_email_with_inline_and_annexes.eml",
                "expected_exit_code": None,
            },
            {
                "filename": "02_email_with_inline_annex_eml.eml",
                "expected_exit_code": None,
            },
            {
                "filename": "03_email_with_false_inline.eml",
                "expected_exit_code": None,
            },
            {
                "filename": "04_email_with_pdf_attachment.eml",
                "expected_exit_code": None,
            },
            {
                "filename": "eml_file_that_does_not_exist.eml",
                "expected_exit_code": 0,
            },
        ]

        for dic in to_test:
            filename = dic["filename"]
            expected_exit_code = dic["expected_exit_code"]
            with patch("sys.argv", ["main.py", "../../config.ini", f"--test_eml={EML_TEST_FILES_PATH}/{filename}"]):
                with patch("imio.email.dms.main.IMAPEmailHandler"):
                    with self.assertRaises(SystemExit) as cm:
                        process_mails()
            self.assertEqual(cm.exception.code, expected_exit_code)

    @patch("imio.email.dms.main.dev_mode", True)
    def test_process_mails_get_eml(self):
        with patch("sys.argv", ["main.py", "../../config.ini", "--get_eml=01"]):
            with patch("imio.email.dms.main.IMAPEmailHandler") as MockIMAPEmailHandler:
                with self.assertRaises(SystemExit) as cm:
                    mock_handler = MockIMAPEmailHandler.return_value
                    mock_handler.get_mail.return_value = get_eml_message("01_email_with_inline_and_annexes.eml")
                    process_mails()
        self.assertIsNone(cm.exception.code)

        with (
            open(os.path.join(EML_TEST_FILES_PATH, "01_email_with_inline_and_annexes.eml"), "r") as original_file,
            open("01.eml", "r") as output_file,
        ):
            # ipdb.set_trace()
            original_mail = Parser(
                email.message_from_file(original_file, policy=email_policy),
                False,
                "01",
            )
            output_mail = Parser(
                email.message_from_file(output_file, policy=email_policy),
                False,
                "01",
                extract=False,
            )
            self.assertEqual(original_mail.parsed_message.body, output_mail.parsed_message.body)
            self.assertEqual(original_mail.parsed_message.headers, output_mail.parsed_message.headers)
            self.assertEqual(original_mail.parsed_message.attachments, output_mail.parsed_message.attachments)

    @patch("imio.email.dms.main.dev_mode", True)
    def test_process_mails_gen_pdf(self):
        with patch("sys.argv", ["main.py", "../../config.ini", "--gen_pdf=01"]):
            with patch("imio.email.dms.main.IMAPEmailHandler") as MockIMAPEmailHandler:
                with self.assertRaises(SystemExit) as cm:
                    mock_handler = MockIMAPEmailHandler.return_value
                    mock_handler.get_mail.return_value = get_eml_message("01_email_with_inline_and_annexes.eml")
                    process_mails()
        self.assertIsNone(cm.exception.code)
        self.assertTrue(Path("/tmp/01.pdf").exists())

    @patch("imio.email.dms.main.dev_mode", True)
    @patch("imio.email.dms.utils.dev_mode", True)
    def test_process_mails_no_option(self):
        with patch("sys.argv", ["main.py", "../../config.ini"]):
            with patch("imio.email.dms.main.IMAPEmailHandler") as MockIMAPEmailHandler:
                with self.assertRaises(SystemExit) as cm:
                    mock_handler = MockIMAPEmailHandler.return_value
                    mock_handler.get_waiting_emails.return_value = [
                        MailData("01", get_eml_message("01_email_with_inline_and_annexes.eml")),
                        MailData("02", get_eml_message("02_email_with_inline_annex_eml.eml")),
                        MailData("03", get_eml_message("03_email_with_false_inline.eml")),
                        MailData("04", get_eml_message("04_email_with_pdf_attachment.eml")),
                    ]
                    process_mails()

        self.assertIsNone(cm.exception.code)

        self.assertTrue(Path("/tmp/01Z999900000001.tar").exists())
        with tarfile.open("/tmp/01Z999900000001.tar", "r") as f:
            self.assertEqual(len(f.getnames()), 4)
            self.assertEqual(
                f.getnames(),
                [
                    "email.pdf",
                    "metadata.json",
                    "/attachments/accuse.odt",
                    "/attachments/Capture du 2016-12-12 10-56-00-(redimensionné).png",
                ],
            )

        self.assertTrue(Path("/tmp/01Z999900000002.tar").exists())
        with tarfile.open("/tmp/01Z999900000002.tar", "r") as f:
            self.assertEqual(len(f.getnames()), 4)
            self.assertEqual(
                f.getnames(),
                [
                    "email.pdf",
                    "metadata.json",
                    "/attachments/texte_simple ééé.txt",
                    "/attachments/Email avec inline et annexe.eml",
                ],
            )

        self.assertTrue(Path("/tmp/01Z999900000003.tar").exists())
        with tarfile.open("/tmp/01Z999900000003.tar", "r") as f:
            self.assertEqual(len(f.getnames()), 4)
            self.assertEqual(
                f.getnames(),
                [
                    "email.pdf",
                    "metadata.json",
                    "/attachments/Erreur 2.jpg",
                    "/attachments/Erreur 1.png",
                ],
            )

    @patch("imio.email.dms.main.dev_mode", True)
    def test_clean_mails_help(self):
        with self.assertRaises(SystemExit):
            docopt(clean_mails.__doc__, argv=["--help"])

    @patch("imio.email.dms.main.dev_mode", True)
    def test_clean_mails_list_only(self):
        emails = {
            "01": get_eml_message("01_email_with_inline_and_annexes.eml"),
            "02": get_eml_message("02_email_with_inline_annex_eml.eml"),
            "03": get_eml_message("03_email_with_false_inline.eml"),
            "04": get_eml_message("04_email_with_pdf_attachment.eml"),
        }
        with (
            patch("sys.argv", ["main.py", "../../config.ini", "--list_only"]),
            patch("imio.email.dms.main.IMAPEmailHandler") as MockIMAPEmailHandler,
            self.assertRaises(SystemExit) as cm,
        ):
            mock_handler = MockIMAPEmailHandler.return_value
            mock_handler.connection.search.return_value = ("OK", ["01 02 03 04"])
            mock_handler.connection.fetch.return_value = ("OK", [b""])
            mock_handler.get_email.side_effect = lambda x: emails[x]
            clean_mails()

    @patch("imio.email.dms.main.Notify._send")
    def test_notify_exception(self, notify_send):
        def assert_msg(msg):
            with open(f"/tmp/{msg['Subject']}.eml", "w") as f:
                f.write(msg.as_string())

        notify_send.side_effect = assert_msg

        msg = get_eml_message("01_email_with_inline_and_annexes.eml")
        parser = Parser(msg, False, "01")
        config = configparser.ConfigParser()
        config.read("../../config.ini")
        try:
            raise Exception("Test exception")
        except Exception as e:
            Notify(parser.message, config, parser.headers).exception("01", e)

        # Assert msg to support
        with open("/tmp/Error handling an email for client 019999.eml", "r") as f:
            msg = email.message_from_file(f)
        self.assertEqual(msg["Subject"], "Error handling an email for client 019999")
        self.assertEqual(msg["From"], "imio.email.dms@imio.be")
        self.assertEqual(msg["To"], "support-docs@imio.be")
        self.assertEqual(
            msg.get_payload(0).get_payload(),
            "\nProblematic mail is attached.\n\nClient ID : 019999\nIMAP login : \n\nmail id : 01\n\nCorresponding exception : <class 'Exception'>\nTest exception\n\n\n\n",
        )
        self.assertEqual(msg.get_payload(1).get_content_type(), "message/rfc822")
        self.assertEqual(msg.get_payload(1)["Content-Disposition"], "inline")

        # Assert msg to client
        with open("/tmp/Erreur dans le traitement du mail transféré vers iA.Docs.eml", "r") as f:
            msg = email.message_from_file(f)
        self.assertEqual(
            msg["Subject"], "=?utf-8?q?Erreur_dans_le_traitement_du_mail_transf=C3=A9r=C3=A9_vers_iA=2EDocs?="
        )
        self.assertEqual(msg["From"], "imio.email.dms@imio.be")
        self.assertEqual(msg["To"], "stephan.mio@mail.be")
        self.assertEqual(msg.get_payload(1).get_content_type(), "message/rfc822")
        self.assertEqual(msg.get_payload(1)["Content-Disposition"], "inline")

    @patch("imio.email.dms.main.Notify._send")
    def test_notify_unsupported_origin(self, notify_send):
        def assert_msg(msg):
            self.assertEqual(msg["Subject"], "Erreur de transfert de votre email dans iA.Docs")
            self.assertEqual(msg["From"], "imio.email.dms@imio.be")
            self.assertEqual(msg["To"], "stephan.perso@mail.be")
            self.assertEqual(msg.get_payload(1).get_content_type(), "message/rfc822")
            self.assertEqual(msg.get_payload(1)["Content-Disposition"], "inline")

        notify_send.side_effect = assert_msg

        msg = get_eml_message("01_email_with_inline_and_annexes.eml")
        parser = Parser(msg, False, "01")
        config = configparser.ConfigParser()
        config.read("../../config.ini")
        Notify(parser.message, config, parser.headers).unsupported_origin()

    @patch("imio.email.dms.main.Notify._send")
    def test_notify_ignored(self, notify_send):
        def assert_msg(msg):
            self.assertEqual(msg["Subject"], "Transfert non autorisé de stephan.mio@mail.be pour le client 019999")
            self.assertEqual(msg["From"], "imio.email.dms@imio.be")
            self.assertEqual(msg["To"], "stephan.mio@mail.be")
            self.assertEqual(msg["Bcc"], "support-docs@imio.be")
            # self.assertEqual(msg["Bcc"], "support-docs@imio.be")
            self.assertEqual(msg.get_payload(1).get_content_type(), "message/rfc822")
            self.assertEqual(msg.get_payload(1)["Content-Disposition"], "inline")

        notify_send.side_effect = assert_msg

        msg = get_eml_message("01_email_with_inline_and_annexes.eml")
        parser = Parser(msg, False, "01")
        config = configparser.ConfigParser()
        config.read("../../config.ini")
        Notify(parser.message, config, parser.headers).ignored("01")

    @patch("imio.email.dms.main.Notify._send")
    def test_notify_result(self, notify_send):
        def assert_msg(msg):
            self.assertEqual(msg["Subject"], "Result of clean_mails for client 019999")
            self.assertEqual(msg["From"], "imio.email.dms@imio.be")
            self.assertEqual(msg["To"], "support-docs@imio.be")
            self.assertIn("Some message", msg.get_payload(0).get_payload())

        notify_send.side_effect = assert_msg

        msg = get_eml_message("01_email_with_inline_and_annexes.eml")
        parser = Parser(msg, False, "01")
        config = configparser.ConfigParser()
        config.read("../../config.ini")
        Notify(parser.message, config, parser.headers).result("Result of clean_mails", "Some message")



# -*- coding: utf-8 -*-

"""
Usage: process_mails FILE [--requeue_errors] [--list_emails=<number>] [--get_eml=<mail_id>] [--gen_pdf=<mail_id>]
                          [--eml_orig] [--reset_flags=<mail_id>] [--test_eml=<path>] [--stats] [--mail_id=<mail_id>]

Arguments:
    FILE         config file

Options:
    -h --help               Show this screen.
    --requeue_errors        Put email in error status back in waiting for processing.
    --list_emails=<number>  List last xx emails.
    --get_eml=<mail_id>     Get eml of original/contained email id.
    --eml_orig              With --get_eml or --test_eml, consider original mail not contained.
    --gen_pdf=<mail_id>     Generate pdf of contained email id.
    --reset_flags=<mail_id> Reset all flags of email id.
    --test_eml=<path>       Test an eml handling.
    --stats                 Get email stats following stats.
    --mail_id=<mail_id>     Use this mail id.
"""
from bs4 import BeautifulSoup
from datetime import datetime
from datetime import timedelta
from docopt import docopt
from email2pdf2 import email2pdf2
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from hashlib import md5
from imio.email.dms import dev_mode
from imio.email.dms import logger
from imio.email.dms.imap import IMAPEmailHandler
from imio.email.dms.imap import MailData
from imio.email.dms.utils import get_next_id
from imio.email.dms.utils import get_reduced_size
from imio.email.dms.utils import get_unique_name
from imio.email.dms.utils import safe_text
from imio.email.dms.utils import save_as_eml
from imio.email.dms.utils import save_attachment  # noqa
from imio.email.dms.utils import set_next_id
from imio.email.parser import email_policy  # noqa
from imio.email.parser.parser import Parser  # noqa
from imio.email.parser.utils import stop  # noqa
from imio.email.parser.utils import structure  # noqa
from imio.pyutils.system import runCommand
from io import BytesIO
from PIL import Image
from PIL import ImageFile
from PIL import ImageOps
from PIL import UnidentifiedImageError
from smtplib import SMTP
from time import sleep
from xml.etree.ElementTree import ParseError

import configparser
import copy
import email
import imaplib
import json
import os
import re
import requests
import six
import sys
import tarfile
import tempfile
import zc.lockfile


try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path  # noqa


dev_infos = {"nid": None}
img_size_limit = 1024
# originally 89478485 => blocks at > 13300 pixels square
Image.MAX_IMAGE_PIXELS = None
# OSError: broken data stream when reading image file
ImageFile.LOAD_TRUNCATED_IMAGES = True
EXIF_ORIENTATION = 0x0112
MAX_SIZE_ATTACH = 19000000


class DmsMetadataError(Exception):
    """The response from the webservice dms_metadata route is not successful"""


class FileUploadError(Exception):
    """The response from the webservice file_upload route is not successful"""


class OperationalError(Exception):
    """The response from the webservice failed due to an OperationalError"""


def check_transferer(sender, pattern):
    if re.match(pattern, sender, re.I):
        return True
    return False


def get_mailbox_infos(config):
    mailbox_infos = config["mailbox"]
    host = str(mailbox_infos["host"])
    port = int(mailbox_infos["port"])
    ssl = mailbox_infos["ssl"] == "true" and True or False
    login = mailbox_infos["login"]
    password = mailbox_infos["pass"]
    return host, port, ssl, login, password


def get_preview_pdf_path(config, mail_id):
    mail_infos = config["mailinfos"]
    output_dir = mail_infos["pdf-output-dir"]
    if isinstance(mail_id, bytes):
        filename = "{0}.pdf".format(mail_id.decode("UTF-8"))
    else:
        filename = "{0}.pdf".format(mail_id)
    return os.path.join(output_dir, filename)


def compress_pdf(original_pdf_content):
    with tempfile.NamedTemporaryFile(mode="wb", delete=False, suffix=".pdf") as input_temp_file:
        input_temp_file.write(original_pdf_content)
        input_temp_file_name = input_temp_file.name

    try:
        # Create a temporary file for output
        with tempfile.NamedTemporaryFile(delete=False, suffix=".pdf") as output_temp_file:
            output_temp_file_name = output_temp_file.name

        # Ghostscript command
        # -dPDFSETTINGS=/screen selects low-resolution output similar to the Acrobat Distiller "Screen Optimized" setting.
        # -dPDFSETTINGS=/ebook selects medium-resolution output similar to the Acrobat Distiller "eBook" setting.
        # -dPDFSETTINGS=/printer selects output similar to the Acrobat Distiller "Print Optimized" setting.
        # -dPDFSETTINGS=/prepress selects output similar to Acrobat Distiller "Prepress Optimized" setting.
        # -dPDFSETTINGS=/default selects output intended to be useful across a wide variety of uses, possibly at the expense of a larger output file.
        gs_command = (
            f"gs -sDEVICE=pdfwrite -dCompatibilityLevel=1.4 -dPDFSETTINGS=/ebook -dNOPAUSE -dQUIET -dBATCH "
            f"-sOutputFile={output_temp_file_name} {input_temp_file_name}"
        )
        _, stderr, returncode = runCommand(gs_command)

        if returncode != 0:
            raise RuntimeError("Ghostscript failed: {}".format("\n".join(stderr)))

        with open(output_temp_file_name, "rb") as f:
            compressed_pdf_content = f.read()

    finally:
        # Clean temporary files
        os.unlink(input_temp_file_name)
        if os.path.exists(output_temp_file_name):
            os.unlink(output_temp_file_name)

    return compressed_pdf_content


def modify_attachments(mail_id, attachments, with_inline=True):
    """Modify parser attachments by reducing images size

    :param mail_id: mail id
    :param attachments: list of attachments
    :param with_inline: keep inline images to reduce size too
    :return: new list of attachments
    """
    new_lst = []
    for dic in attachments:
        # {k: v for k, v in dic.items() if k != 'content'}
        is_inline = False
        # we pass inline image, often used in signature. This image will be in generated pdf
        if dic["type"].startswith("image/") and dic["disp"] == "inline":
            if with_inline:
                is_inline = True
            else:
                if dev_mode:
                    logger.info("{}: skipped inline image '{}' of size {}".format(mail_id, dic["filename"], dic["len"]))
                continue
        if dic["type"].startswith("image/"):
            orient_mod = size_mod = False
            try:
                img = Image.open(BytesIO(dic["content"]))
            except UnidentifiedImageError:
                if not is_inline:
                    new_lst.append(dic)  # kept original image
                continue
            except Image.DecompressionBombError:  # never append because Image.MAX_IMAGE_PIXELS is set to None
                continue
            dic["is_inline"] = is_inline
            try:
                exif = img.getexif()
                orient = exif.get(EXIF_ORIENTATION, 0)
            except ParseError:
                logger.warning(
                    "{}: error getting exif info for image '{}', ignored orientation".format(mail_id, dic["filename"])
                )
                orient = 0
            new_img = img
            # if problem, si ImageMagik use https://github.com/IMIO/appy/blob/master/appy/pod/doc_importers.py#L545
            if not is_inline and orient and orient != 1:
                try:
                    new_img = ImageOps.exif_transpose(img)
                    orient_mod = True
                    if dev_mode:
                        logger.info("{}: reoriented image '{}' from {}".format(mail_id, dic["filename"], orient))
                except Exception:
                    pass
            is_reduced = False
            if is_inline:
                is_reduced, new_size = get_reduced_size(new_img.size, (1000, None))
            elif dic["len"] > 100000:
                is_reduced, new_size = get_reduced_size(new_img.size, (img_size_limit, img_size_limit))
            if is_reduced:
                if dev_mode:
                    logger.info("{}: resized image '{}'".format(mail_id, dic["filename"]))
                # see https://pillow.readthedocs.io/en/stable/handbook/concepts.html#filters
                new_img = new_img.resize(new_size, Image.BICUBIC)
                dic["size"] = new_size
                size_mod = True

            if size_mod or orient_mod:
                new_bytes = BytesIO()
                # save the image in new_bytes
                try:
                    new_img.save(new_bytes, format=img.format, optimize=True, quality=75)
                except ValueError:
                    new_img.save(new_bytes, format=img.format, optimize=True)
                new_content = new_bytes.getvalue()
                new_len = len(new_content)
                if orient_mod or (new_len < dic["len"] and float(new_len / dic["len"]) < 0.9):
                    #                                      more than 10% of difference
                    dic["filename"] = re.sub(r"(\.\w+)$", r"-(redimensionné)\1", dic["filename"])
                    if dev_mode:
                        logger.info(
                            "{}: new image '{}' ({} => {}){}".format(
                                mail_id, dic["filename"], dic["len"], new_len, is_inline and " (inline)" or ""
                            )
                        )
                    dic["len"] = new_len
                    dic["content"] = new_content
                    dic["modified"] = True

        if dic["type"] == "application/pdf":
            new_content = compress_pdf(dic["content"])
            new_len = len(new_content)
            if new_len < dic["len"] and float(new_len / dic["len"]) < 0.9:
                dic["content"] = new_content
                dic["len"] = new_len
                dic["filename"] = re.sub(r"(\.\w+)$", r"-(redimensionné)\1", dic["filename"])
                dic["modified"] = True
                if dev_mode:
                    logger.info("{}: new pdf '{}' ({} => {})".format(mail_id, dic["filename"], dic["len"], new_len))

        new_lst.append(dic)
    return new_lst


def resize_inline_images(mail_id, message, attachments):
    new_message = copy.deepcopy(message)
    size_pattern = r'(size=["\'])([^"\']*)(["\'])'
    cids = {}
    # replace inline image content with reduced one
    for at in attachments:
        if not (at["type"].startswith("image/") and at["disp"] == "inline" and at.get("modified")):
            continue
        part = email2pdf2.find_part_by_content_id(new_message, at["cid"])
        if not part:
            continue
        cids[at["cid"].strip("<>")] = {"sz": at.get("size"), "used": False}
        disposition = part.get("Content-Disposition")
        if disposition and "size=" in disposition:
            disposition = re.sub(size_pattern, lambda m: f"{m.group(1)}{at['len']}{m.group(3)}", disposition)
        # Replace the image content in the new message
        part.set_content(
            at["content"],
            maintype=part.get_content_maintype(),
            subtype=part.get_content_subtype(),
            disposition=disposition,
            cte="base64",
            cid=part.get("Content-ID"),
        )
        logger.info(f"{mail_id}: replaced inline {at['cid']} ({at['filename']})")
    if not cids:
        return message
    # save_as_eml("/tmp/a.eml", new_message)
    # resize inline images in html part
    for part in new_message.walk():
        if part.get_content_type() != "text/html":
            continue
        html_part = part
        changes = set()
        soup = BeautifulSoup(html_part.get_content(), "html.parser")
        for img_tag in soup.find_all("img"):
            img_cid = img_tag.get("src", "").replace("cid:", "").strip()
            if not img_cid or img_cid not in cids:
                continue
            cids[img_cid]["used"] = True
            style_dic = {}
            if "style" in img_tag.attrs:
                styles = img_tag["style"].split(";")
                for style in styles:
                    if ":" in style:
                        key, value = style.split(":", 1)
                        style_dic[key.strip()] = value.strip()
            style_dic["max-width"] = "100%"
            style_dic["height"] = "auto"
            current_width = style_dic.get("width", "auto")
            if (
                current_width != "auto"
                and "px" in current_width
                and int(current_width.replace("px", "")) > cids[img_cid]["sz"][0]
            ):
                style_dic["width"] = "auto"
            else:
                style_dic["width"] = current_width

            img_tag["style"] = "; ".join(f"{k}: {v}" for k, v in style_dic.items())
            changes.add(img_cid)
        if changes:
            html_part.set_content(soup.prettify(), subtype="html")
            c_str = ", ".join(changes)
            logger.info(f"{mail_id}: resized inline images cids '{c_str}' in html part")
    for cid in cids:
        if not cids[cid]["used"]:
            logger.warning("{}: inline image '{}' not found".format(mail_id, cid))
    return new_message


def post_with_retries(url, auth, action, mail_id, json_data=None, files=None, retries=5, delay=20):
    for attempt in range(1, retries + 1):
        try:
            response = requests.post(url, auth=auth, json=json_data, files=files)
            # can simulate an empty response when webservice communication problems
            # response._content = b""
            if not response.content:
                raise requests.exceptions.RequestException("Empty response")
            response.raise_for_status()  # Raise an HTTPError for bad responses
            req_content = json.loads(response.content)
            # can be simulated by stopping postgresql
            if "error" in req_content and "(OperationalError) " in req_content["error"]:
                raise OperationalError(req_content["error"])
            return req_content
        except (OperationalError, requests.exceptions.RequestException) as e:
            if attempt < retries:
                sleep(delay)
                logger.info(f"{mail_id}: failed attempt {attempt} to {action}: '{e}'")
            else:
                raise e


def send_to_ws(config, headers, main_file_path, attachments, mail_id):
    ws = config["webservice"]
    next_id, client_id = get_next_id(config, dev_infos)
    external_id = "{0}{1:08d}".format(client_id, next_id)

    tar_path = Path("/tmp") / "{}.tar".format(external_id)
    with tarfile.open(str(tar_path), "w") as tar:
        # 1) email pdf printout or eml file
        with Path(main_file_path).open("rb") as f:
            mf_contents = f.read()
        basename, ext = os.path.splitext(main_file_path)
        mf_info = tarfile.TarInfo(name="email{}".format(ext))
        mf_info.size = len(mf_contents)
        tar.addfile(tarinfo=mf_info, fileobj=BytesIO(mf_contents))

        # 2) metadata.json
        metadata_contents = json.dumps(headers).encode("utf8") if six.PY3 else json.dumps(headers)
        metadata_info = tarfile.TarInfo(name="metadata.json")
        metadata_info.size = len(metadata_contents)
        tar.addfile(tarinfo=metadata_info, fileobj=BytesIO(metadata_contents))

        # 3) every attachment file
        files = []
        for attachment in attachments:
            if attachment.get("is_inline"):
                continue
            attachment_contents = attachment["content"]
            attachment_info = tarfile.TarInfo(
                name="/attachments/{}".format(get_unique_name(attachment["filename"], files))
            )
            attachment_info.size = len(attachment_contents)
            tar.addfile(tarinfo=attachment_info, fileobj=BytesIO(attachment_contents))
    if dev_mode:
        logger.info("tar file '{}' created".format(tar_path))
    else:  # we send to the ws
        tar_content = tar_path.read_bytes()
        now = datetime.now()
        metadata = {
            "external_id": external_id,
            "client_id": client_id,
            "scan_date": now.strftime("%Y-%m-%d"),
            "scan_hour": now.strftime("%H:%M:%S"),
            "user": "testuser",
            "pc": "pc-scan01",
            "creator": "scanner",
            "filesize": len(tar_content),
            "filename": tar_path.name,
            "filemd5": md5(tar_content).hexdigest(),
        }

        auth = (ws["login"], ws["pass"])
        proto = ws["port"] == "443" and "https" or "http"
        metadata_url = "{proto}://{ws[host]}:{ws[port]}/dms_metadata/{client_id}/{ws[version]}".format(
            proto=proto,
            ws=ws,
            client_id=client_id,
        )
        metadata_req_content = post_with_retries(metadata_url, auth, "post metadata", mail_id, json_data=metadata)
        # {'message': 'Well done', 'external_id': '05Z507000024176', 'id': 2557054, 'success': True}
        if not metadata_req_content["success"] or "id" not in metadata_req_content:
            msg = "mail_id: {}, code: '{}', error: '{}', metadata: '{}'".format(
                mail_id, metadata_req_content["error_code"], metadata_req_content["error"], metadata
            ).encode("utf8")
            raise DmsMetadataError(msg)
        response_id = metadata_req_content["id"]

        upload_url = "{proto}://{ws[host]}:{ws[port]}/file_upload/{ws[version]}/{id}".format(
            proto=proto, ws=ws, id=response_id
        )
        files = {"filedata": ("archive.tar", tar_content, "application/tar", {"Expires": "0"})}
        upload_req_content = post_with_retries(upload_url, auth, "upload file", mail_id, files=files, retries=5)
        if not upload_req_content["success"]:
            msg = "mail_id: {}, code: '{}', error: '{}'".format(
                mail_id,
                upload_req_content["error_code"],
                upload_req_content.get("error") or upload_req_content["message"],
            ).encode("utf8")
            raise FileUploadError(msg)

        set_next_id(config, next_id)


def process_mails():
    arguments = docopt(__doc__)
    config = configparser.ConfigParser()
    config_file = arguments["FILE"]
    config.read(config_file)

    host, port, ssl, login, password = get_mailbox_infos(config)
    counter_dir = Path(config["webservice"]["counter_dir"])
    counter_dir.mkdir(exist_ok=True)
    lock_filepath = counter_dir / "lock_{0}".format(config["webservice"]["client_id"])
    lock = zc.lockfile.LockFile(lock_filepath.as_posix())

    handler = IMAPEmailHandler()
    handler.connect(host, port, ssl, login, password)

    if arguments.get("--requeue_errors"):
        amount = handler.reset_errors()
        logger.info("{} emails in error were put back in waiting state".format(amount))
        handler.disconnect()
        lock.close()
        sys.exit()
    elif arguments.get("--list_emails"):
        handler.list_last_emails(nb=int(arguments.get("--list_emails")))
        # import ipdb; ipdb.set_trace()
        # handler.mark_reset_error('58')
        # handler.mark_reset_ignored('77')
        # handler.mark_mail_as_imported('594')
        # res, data = handler.connection.search(None, 'SUBJECT "FAIGNART MARION"')
        # for mail_id in data[0].split():
        #      omail = handler.get_mail(mail_id)
        #      parser = Parser(omail, dev_mode, mail_id)
        #      headers = parser.headers
        #      amail = parser.message
        #      parsed = MailParser(omail)
        #     logger.info(headers['Subject'])
        handler.disconnect()
        lock.close()
        sys.exit()
    elif arguments.get("--get_eml"):
        mail_id = arguments["--get_eml"]
        if not mail_id:
            stop("Error: you must give an email id (--get_eml=25 by example)", logger)
        try:
            mail = handler.get_mail(mail_id)
            parsed = Parser(mail, dev_mode, mail_id)
            logger.info(parsed.headers)
            message = parsed.message
            # structure(message)
            filename = "{}.eml".format(mail_id)
            if login:
                filename = "{}_{}".format(login, filename)
            if arguments.get("--eml_orig"):
                message = parsed.initial_message
                filename = filename.replace(".eml", "_o.eml")
            logger.info("Writing {} file".format(filename))
            # o_attachments = parsed.attachments(False, set())
            save_as_eml(filename, message)
        except Exception as e:
            logger.error(e, exc_info=True)
            Notify(mail, config, None).exception(mail_id, e)
            if not dev_mode:
                handler.mark_mail_as_error(mail_id)
        handler.disconnect()
        lock.close()
        sys.exit()
    elif arguments.get("--gen_pdf"):
        mail_id = arguments["--gen_pdf"]
        if not mail_id:
            stop("Error: you must give an email id (--gen_pdf=25 by example)", logger)
        mail = handler.get_mail(mail_id)
        parsed = Parser(mail, dev_mode, mail_id)
        logger.info(parsed.headers)
        pdf_path = get_preview_pdf_path(config, mail_id.encode("utf8"))
        logger.info("Generating {} file".format(pdf_path))
        attachments = modify_attachments(mail_id, parsed.attachments)
        message = resize_inline_images(mail_id, parsed.message, attachments)
        parsed.generate_pdf(pdf_path, message=message)
        handler.disconnect()
        lock.close()
        sys.exit()
    elif arguments.get("--reset_flags"):
        mail_id = arguments["--reset_flags"]
        if not mail_id:
            stop("Error: you must give an email id (--reset_flags=25 by example)", logger)
        # handler.mark_mail_as_error(mail_id)
        # handler.mark_mail_as_imported(mail_id)
        handler.mark_reset_all(mail_id)
        handler.disconnect()
        lock.close()
        sys.exit()
    elif arguments.get("--test_eml"):
        handler.disconnect()
        eml_path = arguments["--test_eml"]
        if not eml_path or not os.path.exists(eml_path):
            stop(
                "Error: you must give an existing eml path '{}' (--test_eml=123.eml by example)".format(eml_path),
                logger,
            )
        if not dev_mode:
            stop("Error: You must activate dev mode to test an eml file", logger)
        with open(eml_path) as fp:
            mail = email.message_from_file(fp, policy=email_policy)
        mail_id = os.path.splitext(os.path.basename(eml_path))[0]
        if not arguments.get("--eml_orig"):
            mail.__setitem__("X-Forwarded-For", "0.0.0.0")  # to be considered as main mail
        parser = Parser(mail, dev_mode, "")
        headers = parser.headers
        # [{k: v for k, v in at.items() if k != 'content'} for at in parser.attachments]
        attachments = modify_attachments(mail_id, parser.attachments)
        # save_attachment(config["mailinfos"]["pdf-output-dir"], attachments[0])
        try:
            main_file_path = get_preview_pdf_path(config, mail_id)
            logger.info("pdf file {}".format(main_file_path))
            message = resize_inline_images(mail_id, parser.message, attachments)
            parser.generate_pdf(main_file_path, message=message)
            if dev_mode:
                save_as_eml(main_file_path.replace(".pdf", ".eml"), message)
        except Exception:
            logger.error("Error generating pdf file", exc_info=True)
            main_file_path = main_file_path.replace(".pdf", ".eml")
            save_as_eml(main_file_path, parser.message)
        send_to_ws(config, headers, main_file_path, attachments, mail_id)
        lock.close()
        sys.exit()
    elif arguments.get("--stats"):
        logger.info("Started at {}".format(datetime.now()))
        stats = handler.stats()
        logger.info("Total mails: {}".format(stats.pop("tot")))
        for flag in sorted(stats["flags"]):
            logger.info("Flag '{}' => {}".format(flag, stats["flags"][flag]))
        handler.disconnect()
        lock.close()
        logger.info("Ended at {}".format(datetime.now()))
        sys.exit()

    imported = errors = unsupported = ignored = total = 0
    if arguments.get("--mail_id"):
        mail_id = arguments["--mail_id"]
        if not mail_id:
            stop("Error: you must give an email id (--mail_id=25 by example)", logger)
        mail = handler.get_mail(mail_id)
        if not mail:
            stop("Error: no mail found for id {}".format(mail_id), logger)
        emails = [MailData(mail_id, mail)]
    else:
        emails = handler.get_waiting_emails()
    for mail_info in emails:
        total += 1
        mail_id = mail_info.id
        mail = mail_info.mail
        main_file_path = get_preview_pdf_path(config, mail_id)
        parser = None
        try:
            parser = Parser(mail, dev_mode, mail_id)
            headers = parser.headers
            if parser.origin == "Generic inbox":
                if not dev_mode:
                    handler.mark_mail_as_unsupported(mail_id)
                unsupported += 1
                try:
                    Notify(mail, config, headers).unsupported_origin()
                except Exception:  # better to continue than advise user
                    pass
                continue
            # we check if the pushing agent has a permitted email format
            if "Agent" in headers and not check_transferer(
                headers["Agent"][0][1], config["mailinfos"].get("sender-pattern", ".+")
            ):
                if not dev_mode:
                    handler.mark_mail_as_ignored(mail_id)
                # logger.error('Rejecting {}: {}'.format(headers['Agent'][0][1], headers['Subject']))
                ignored += 1
                try:
                    Notify(mail, config, headers).ignored(mail_id)
                except Exception:  # better to continue than advise user
                    pass
                continue
            # logger.info('Accepting {}: {}'.format(headers['Agent'][0][1], headers['Subject']))
            try:
                attachments = modify_attachments(mail_id, parser.attachments)
            except Exception:
                logger.error("Error modifying attachments", exc_info=True)
                attachments = parser.attachments
            try:
                message = resize_inline_images(mail_id, parser.message, attachments)
            except Exception:
                logger.error("Error resizing inline images", exc_info=True)
                message = parser.message
            try:
                parser.generate_pdf(main_file_path, message=message)
            except Exception:
                logger.error("Error generating pdf file", exc_info=True)
                # if 'XDG_SESSION_TYPE=wayland' not in str(pdf_exc):
                main_file_path = main_file_path.replace(".pdf", ".eml")
                save_as_eml(main_file_path, parser.message)
            send_to_ws(config, headers, main_file_path, attachments, mail_id)
            if not dev_mode:
                handler.mark_mail_as_imported(mail_id)
            imported += 1
        except Exception as e:
            logger.error(e, exc_info=True)
            try:
                # check parser and parser.headers
                Notify(mail, config, parser.headers).exception(mail_id, e)
            except Exception:
                Notify(mail, config, None).exception(mail_id, e)
            if not dev_mode:
                handler.mark_mail_as_error(mail_id)
            errors += 1

    if total:
        logger.info(
            "Treated {} emails: {} imported. {} unsupported. {} in error. {} ignored.".format(
                total, imported, unsupported, errors, ignored
            )
        )
    else:
        logger.info("Treated no email.")
    handler.disconnect()
    lock.close()
    sys.exit()


def clean_mails():
    """Clean mails from imap box.

    Usage: clean_mails FILE [-h] [--kept_days=<number>] [--ignored_too] [--list_only]

    Arguments:
        FILE         config file

    Options:
        -h --help               Show this screen.
        --kept_days=<number>    Days to keep [default: 30]
        --ignored_too           Get also not imported emails
        --list_only             Only list related emails, do not delete
    """
    arguments = docopt(clean_mails.__doc__)
    config = configparser.ConfigParser()
    config.read(arguments["FILE"])
    days = int(arguments["--kept_days"])
    doit = not arguments["--list_only"]
    host, port, ssl, login, password = get_mailbox_infos(config)
    handler = IMAPEmailHandler()
    handler.connect(host, port, ssl, login, password)
    before_date = (datetime.now() - timedelta(days)).strftime("%d-%b-%Y")  # date string 01-Jan-2021
    # before_date = '01-Jun-2021'
    res, data = handler.connection.search(None, "(BEFORE {0})".format(before_date))
    if res != "OK":
        logger.error("Unable to fetch mails before '{}'".format(before_date))
        handler.disconnect()
        sys.exit()
    deleted = ignored = error = 0
    mail_ids = data[0].split()
    mail_ids_len = len(mail_ids)
    out = ["Get '{}' emails older than '{}'".format(mail_ids_len, before_date)]
    logger.info("Get '{}' emails older than '{}'".format(mail_ids_len, before_date))
    # sys.exit()
    for mail_id in mail_ids:
        res, flags_data = handler.connection.fetch(mail_id, "(FLAGS)")
        if res != "OK":
            logger.error("Unable to fetch flags for mail {0}".format(mail_id))
            error += 1
            continue
        flags = imaplib.ParseFlags(flags_data[0])
        if not arguments["--ignored_too"] and b"imported" not in flags:
            ignored += 1
            continue
        mail = handler.get_mail(mail_id)
        if not mail:
            error += 1
            continue
        parser = Parser(mail, dev_mode, mail_id)
        logger.info("{}: '{}'".format(mail_id, parser.headers["Subject"]))
        out.append("{}: '{}'".format(mail_id, parser.headers["Subject"]))
        if doit:
            handler.connection.store(mail_id, "+FLAGS", "\\Deleted")
        deleted += 1
    if deleted:
        logger.info("Get '{}' emails older than '{}'".format(mail_ids_len, before_date))
        if doit:
            res, data = handler.connection.expunge()
            if res != "OK":
                out.append("ERROR: Unable to delete mails !!")
                logger.error("Unable to delete mails")
    handler.disconnect()
    out.append(
        "{} emails have been deleted. {} emails are ignored. {} emails have caused an error.".format(
            deleted, ignored, error
        )
    )
    logger.info(
        "{} emails have been deleted. {} emails are ignored. {} emails have caused an error.".format(
            deleted, ignored, error
        )
    )
    Notify(None, config, None).result("Result of clean_mails", "\n".join(out))
    sys.exit()


class Notify:
    ERROR_MAIL_SUPPORT = """
Problematic mail is attached.\n
Client ID : {0}
IMAP login : {1}\n
mail id : {2}\n
Corresponding exception : {3}
{4}\n
{additional}\n
"""

    ERROR_MAIL_AGENT = """
Cher utilisateur d'iA.Docs,

Vous avez transféré un email vers iA.Docs (sujet: « {0} »).
Malheureusement, une erreur est survenue lors du traitement de cet email et il n'a pas été intégré dans l'application.\n
Nous avons été averti de l'erreur et allons y regarder.\n
Il ne sert à rien d'envoyer à nouveau l'email.\n
Cordialement.\n
{additional}\n
"""

    UNSUPPORTED_ORIGIN_EMAIL = """
Cher utilisateur d'iA.Docs,

Le transfert de l'email attaché (sujet: « {0} ») a été rejeté car il n'a pas été transféré correctement.\n
Veuillez refaire le transfert du mail original en transférant "en tant que pièce jointe".\n
Si vous utilisez Microsoft Outlook:\n
- Dans le ruban, cliquez sur la flèche du ménu déroulant située sur le bouton de transfert\n
- Choisissez le transfert en tant que pièce jointe\n
- Envoyez le mail sans rien compléter d'autre à l'adresse prévue pour iA.Docs.\n
\n
Si vous utilisez Mozilla Thunderbird:\n
- Faites un clic droit sur l'email pour ouvrir le menu contextuel\n
- Sélectionnez "Transférer au format" > "Pièce jointe".\n
- Envoyez le mail sans rien compléter d'autre à l'adresse prévue pour iA.Docs.\n
\n
Cordialement.\n
{additional}\n
"""

    IGNORED_MAIL = """
Bonjour,
Votre adresse email {3} n'est pas autorisée à transférer un email vers iA.docs.
Si cette action est justifiée, veuillez prendre contact avec votre référent interne.\n
Le mail concerné est en pièce jointe (sujet: « {4} »).\n
Client ID : {0}
IMAP login : {1}
mail id : {2}
pattern : "caché"

Cordialement.\n
{additional}\n
"""

    RESULT_MAIL = """
Client ID : {0}
IMAP login : {1}\n
{2}\n
"""

    def __init__(self, mail, config, headers):
        self.mail = mail
        self.config = config
        self.smtp_infos = self.config["smtp"]
        self.headers = headers

    def _set_message(self, msg, unformatted_message, format_args):
        mail_string = self.mail.as_string()
        len_ok = True
        additional = ""

        if len(mail_string) > MAX_SIZE_ATTACH:
            len_ok = False
            additional = "La pièce jointe est trop grosse: on ne sait pas l'envoyer par mail !"

        main_text = MIMEText(unformatted_message.format(*format_args, additional=additional), "plain")
        msg.attach(main_text)
        if len_ok:
            attachment = MIMEBase("message", "rfc822")
            attachment.set_payload(mail_string, "utf8")
            attachment.add_header("Content-Disposition", "inline")
            msg.attach(attachment)

        return msg

    def _send(self, msg):
        smtp = SMTP(str(self.smtp_infos["host"]), int(self.smtp_infos["port"]))
        smtp.send_message(msg)
        smtp.quit()

    def exception(self, mail_id, error):
        def to_support():
            msg = MIMEMultipart()
            msg["Subject"] = "Error handling an email for client {}".format(self.config["webservice"]["client_id"])
            msg["From"] = self.smtp_infos["sender"]
            msg["To"] = self.smtp_infos["recipient"]

            error_msg = error
            if hasattr(error, "message"):
                error_msg = safe_text(error.message)
            elif hasattr(error, "reason"):
                try:
                    error_msg = "'{}', {}, {}, {}".format(error.reason, error.start, error.end, error.object)
                except Exception:
                    error_msg = error.reason

            msg = self._set_message(
                msg,
                self.ERROR_MAIL_SUPPORT,
                format_args=(
                    self.config["webservice"]["client_id"],
                    self.config["mailbox"]["login"],
                    mail_id,
                    error.__class__,
                    error_msg,
                ),
            )
            self._send(msg)

        def to_agent():
            msg = MIMEMultipart()
            msg["Subject"] = "Erreur dans le traitement du mail transféré vers iA.Docs"
            msg["From"] = self.smtp_infos["sender"]
            msg["To"] = self.headers["Agent"][0][1]
            msg = self._set_message(msg, self.ERROR_MAIL_AGENT, format_args=(self.headers["Subject"],))
            self._send(msg)

        to_support()
        if self.headers and "Agent" in self.headers:
            to_agent()

    def unsupported_origin(self):
        msg = MIMEMultipart()
        msg["Subject"] = "Erreur de transfert de votre email dans iA.Docs"
        msg["From"] = self.smtp_infos["sender"]
        msg["To"] = self.headers["From"][0][1]
        msg = self._set_message(msg, self.UNSUPPORTED_ORIGIN_EMAIL, format_args=(self.headers["Subject"],))
        self._send(msg)

    def ignored(self, mail_id):
        msg = MIMEMultipart()
        msg["Subject"] = "Transfert non autorisé de {} pour le client {}".format(
            self.headers["Agent"][0][1], self.config["webservice"]["client_id"]
        )
        msg["From"] = self.smtp_infos["sender"]
        msg["To"] = self.headers["Agent"][0][1]
        msg["Bcc"] = self.smtp_infos["recipient"]
        msg = self._set_message(
            msg,
            self.IGNORED_MAIL,
            format_args=(
                self.config["webservice"]["client_id"],
                self.config["mailbox"]["login"],
                mail_id,
                self.headers["Agent"][0][1],
                self.headers["Subject"],
            ),
        )
        self._send(msg)

    def result(self, subject, message):
        msg = MIMEMultipart()
        msg["Subject"] = "{} for client {}".format(subject, self.config["webservice"]["client_id"])
        msg["From"] = self.smtp_infos["sender"]
        msg["To"] = self.smtp_infos["recipient"]
        main_text = MIMEText(
            self.RESULT_MAIL.format(self.config["webservice"]["client_id"], self.config["mailbox"]["login"], message),
            "plain",
        )
        msg.attach(main_text)
        self._send(msg)
